/*
 * Copyright (C) 2010 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Mikkel Kamstrup Erlandsen <mikkel.kamstrup@canonical.com>
 *
 */

using Zeitgeist;
using Gee;
using Dee;
using Unity.Place;

namespace Unity.FilesPlace.Utils {

    const string icon_attribs = FILE_ATTRIBUTE_PREVIEW_ICON + "," +
                                FILE_ATTRIBUTE_STANDARD_ICON + "," +
                                FILE_ATTRIBUTE_THUMBNAIL_PATH;

    const string file_attribs = FILE_ATTRIBUTE_STANDARD_NAME + "," +
                                FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME + "," +
                                FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE + "," +
                                FILE_ATTRIBUTE_STANDARD_IS_HIDDEN + "," +
                                FILE_ATTRIBUTE_STANDARD_IS_BACKUP + "," +
                                FILE_ATTRIBUTE_TIME_MODIFIED + "," +
                                FILE_ATTRIBUTE_TIME_ACCESS;
    
    const string all_attribs = icon_attribs + "," + file_attribs;

	  /* Synchronous method to query GIO for a good icon for a uri/mimetype */
    public string get_icon_for_uri (string uri, string mimetype)
    {
      try {

        File f = File.new_for_uri (uri);
        FileInfo info = f.query_info (icon_attribs,
                                      FileQueryInfoFlags.NONE,
                                      null);
        return check_icon_string (uri, mimetype, info);
      } catch (GLib.Error e) {
        /* We failed to probe the icon. Try looking up by mimetype instead */
        Icon icon2 = g_content_type_get_icon (mimetype);
        return icon2.to_string ();
      }
    }

    /* Async method to query GIO for a good icon for a uri/mimetype */
    public async string get_icon_for_uri_async (string uri, string mimetype)
    {
      try {

        File f = File.new_for_uri (uri);
        FileInfo info = yield f.query_info_async (icon_attribs,
                                                  FileQueryInfoFlags.NONE,
                                                  Priority.LOW,
                                                  null);
        return check_icon_string (uri, mimetype, info);
      } catch (GLib.Error e) {
        /* We failed to probe the icon. Try looking up by mimetype instead */
        Icon icon2 = g_content_type_get_icon (mimetype);
        return icon2.to_string ();
      }
    }

    public string check_icon_string (string uri, string mimetype, FileInfo info)
    {
      Icon icon = info.get_icon();
      string thumbnail_path = info.get_attribute_byte_string (FILE_ATTRIBUTE_THUMBNAIL_PATH);

      if (thumbnail_path == null)
        return icon.to_string();
      else
        return new FileIcon(File.new_for_path (thumbnail_path)).to_string ();
    }
    
  public int cmp_file_info_by_mtime (FileInfo info1, FileInfo info2)
  {
    TimeVal tv1, tv2;
    info1.get_modification_time (out tv1);
    info2.get_modification_time (out tv2);
    long cmp = tv1.tv_sec - tv2.tv_sec;
    return cmp < 0 ? 1 : ( cmp > 0 ? -1 : 0);
  }

    public string get_month_name (DateMonth month)
    {
      switch (month)
        {
          case DateMonth.BAD_MONTH:
            return _("Invalid Month");
          case DateMonth.JANUARY:
            return _("January");
          case DateMonth.FEBRUARY:
            return _("February");
          case DateMonth.MARCH:
            return _("March");
          case DateMonth.APRIL:
            return _("April");
          case DateMonth.MAY:
            return _("May");
          case DateMonth.JUNE:
            return _("June");
          case DateMonth.JULY:
            return _("July");
          case DateMonth.AUGUST:
            return _("August");
          case DateMonth.SEPTEMBER:
            return _("September");
          case DateMonth.OCTOBER:
            return _("October");
          case DateMonth.NOVEMBER:
            return _("November");
          case DateMonth.DECEMBER:
            return _("December");
        }
        
      return "Internal Error";
    }
    
    /* Calculate the time group of the event returning the group id
     * of the group @event belongs to. It may be necessary for this method
     * to add a row to @groups_model if the group is one of the variadic
     * types such as */
    public uint get_time_group (Zeitgeist.Event event, Dee.Model groups_model)
    {
      // FIXME: We need to calculate bounds of previous midnight, week, month etc
      //        This impl is just a quick hack to have something to show
      
      var t = event.get_timestamp ();
      var now = Timestamp.now ();
      
      /* Set limit to the previous midnight */
      var limit = Timestamp.prev_midnight (now);
      if (t > limit)
        return Group.TODAY;
      else if (t > limit - Zeitgeist.Timestamp.DAY)
        return Group.YESTERDAY;
      
      /* Set limit to start of calendar week (monday) */
      var date = Date();
      Timestamp.to_date (now, out date);
      limit = limit - (date.get_weekday() - 1)*Timestamp.DAY;
      if (t > limit)
        return Group.THIS_WEEK;
      else if (t > limit - Zeitgeist.Timestamp.WEEK)
        return Group.LAST_WEEK;
      
      /* Set limit to start of current calendar month */
      date.set_day (1);
      limit = Timestamp.from_date (date);
      if (t > limit)
        return Group.THIS_MONTH;
      
      /* Set limit to start of previous calendar month */
      date.subtract_months (1);
      limit = Timestamp.from_date (date);
      if (t > now - 6*30*Zeitgeist.Timestamp.DAY)
        return Group.PAST_SIX_MONTHS;
      
      /* Set limit to 1st pf January in the current year. We need to 
       * reset 'date' because date.subtract_months(1) above might have
       * turned a year */
      Timestamp.to_date (now, out date);
      date.set_month (DateMonth.JANUARY);
      date.set_day (1);
      limit = Timestamp.from_date (date);
      if (t > limit)
        return Group.THIS_YEAR;
      else
        return Group.LAST_YEAR;
    }
  
  
    /**
     * Asynchrnously list the contents of a directory given by a path,
     * and return a list of FileInfos filled with the attributes specified
     * by Utils.all_attribs
     */
    public async SList<FileInfo> list_dir (File folder) throws Error
    {
	  var result = new SList<FileInfo> ();
      var e = yield folder.enumerate_children_async (Utils.all_attribs,
                                                     0, Priority.DEFAULT,
                                                     null);

      while (true) {
        var file_infos = yield e.next_files_async (10, Priority.DEFAULT,
                                                   null);
        if (file_infos == null)
          break;

        foreach (var info in file_infos) {
          result.prepend (info);
        }
      }
      
      result.reverse ();
      return result;  
    }
    
  /* Run through @filter_set and remove any row from the result set's model
   * in which URI is not in @valid_uris */
  public void apply_uri_filter (Set<string> valid_uris,
                                Dee.ResultSet filter_set)
  {
    var model = filter_set.get_model ();
    uint n_removed = 0;
    string uri;
    
    /* Anything in filter_set that is not in the event_uris set, is removed
     * from the results_model */
    foreach (var row in filter_set)
    {
      uri = model.get_string (row, ResultsColumn.URI);
      if (!(uri in valid_uris))
        {
          model.remove (row);
          n_removed++;
          debug ("Removed: %s", uri);
        }
    }
    
    if (n_removed > 0)
      debug ("Removed %u rows from result set", n_removed);
  }
  
  /* Returns true iff new_search is a query that simply restricts the
   * result set of old_query */
  public bool check_is_filter_search (Search new_search, Search? old_search)
  {
    bool is_filter_search = false;      
    if (old_search != null)
      {
        string previous_search_string = old_search.get_search_string();
        if (previous_search_string != null && previous_search_string != "")
          is_filter_search = new_search.get_search_string().has_prefix (
                                                      previous_search_string);
      }
    
    return is_filter_search;
  }
  
  /* Extract a hash set of all subject uris in a Zeitgeist.ResultSet */
  public Set<string> get_uri_set (Zeitgeist.ResultSet results)
  {
    Set<string> uris = new HashSet<string> ();
    
    foreach (var ev in results)
      {
        for (int i = 0; i  < ev.num_subjects(); i++)
          uris.add (ev.get_subject (i).get_uri ());
      }
    
    return uris;
  }
    
} /* namespace */
